@@ -21,6 +21,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include "nm-glib-compat.h"
 #include "nm-gsm-device.h"
 #include "nm-device-interface.h"
 #include "nm-device-private.h"
@@ -49,6 +50,9 @@ typedef struct {
 
 	guint reg_tries;
 	guint init_tries;
+
+	gboolean needs_cgreg;
+	gboolean checked_cgmm;
 } NMGsmDevicePrivate;
 
 enum {
@@ -66,10 +70,21 @@ enum {
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
+/* Various possible init sequences */
+const gchar *modem_init_sequences[] = {
+	"ATZ E0 V1 X4 &C1 +FCLASS=0",
+	"ATZ E0 V1 &C1",
+	"AT&F E0 V1 X4 &C1 +FCLASS=0",
+	"AT&F E0 V1 &C1",
+	"AT&F E0 V1",
+	"\rAT&F E0 V1 X4 &C1 +CREG=0 +FCLASS=0",  /* USBModem by MobileStream for Palm */
+	NULL
+};
+
 static void enter_pin (NMGsmDevice *device, NMGsmSecret secret_type, gboolean retry);
 static void manual_registration (NMGsmDevice *device);
 static void automatic_registration (NMGsmDevice *device);
-static void init_modem (NMSerialDevice *device, gpointer user_data);
+static void init_modem (NMSerialDevice *device);
 
 NMGsmDevice *
 nm_gsm_device_new (const char *udi,
@@ -354,7 +369,7 @@ schedule_automatic_registration_again (NMGsmDevice *self)
 	if (priv->pending_id)
 		g_source_remove (priv->pending_id);
 
-	priv->pending_id = g_timeout_add (1000, automatic_registration_again, self);
+	priv->pending_id = g_timeout_add_seconds (1, automatic_registration_again, self);
 }
 
 static void
@@ -428,10 +443,15 @@ automatic_registration_response (NMSerialDevice *device,
 static void
 automatic_registration (NMGsmDevice *device)
 {
-	const char *responses[] = { "+CREG: 0,0", "+CREG: 0,1", "+CREG: 0,2", "+CREG: 0,3", "+CREG: 0,5", NULL };
+	NMGsmDevicePrivate *priv = NM_GSM_DEVICE_GET_PRIVATE (device);
+	const char *creg_responses[] = { "+CREG: 0,0", "+CREG: 0,1", "+CREG: 0,2", "+CREG: 0,3", "+CREG: 0,5", NULL };
+	const char *cgreg_responses[] = { "+CGREG: 0,0", "+CGREG: 0,1", "+CGREG: 0,2", "+CGREG: 0,3", "+CGREG: 0,5", NULL };
 	const char *terminators[] = { "OK", "ERROR", "ERR", NULL };
 
-	modem_wait_for_reply (device, "AT+CREG?", 15, responses, terminators, automatic_registration_response, NULL);
+	if (priv->needs_cgreg)
+		modem_wait_for_reply (device, "AT+CGREG?", 15, cgreg_responses, terminators, automatic_registration_response, NULL);
+	else
+		modem_wait_for_reply (device, "AT+CREG?", 15, creg_responses, terminators, automatic_registration_response, NULL);
 }
 
 static void
@@ -449,14 +469,43 @@ do_register (NMGsmDevice *device)
 		automatic_registration (device);
 }
 
+static void
+get_model_done (NMSerialDevice *device,
+                int reply_index,
+                const char *reply,
+                gpointer user_data)
+{
+	NMGsmDevicePrivate *priv = NM_GSM_DEVICE_GET_PRIVATE (device);
+
+	priv->checked_cgmm = TRUE;
+
+	switch (reply_index) {
+	case 0:
+		priv->needs_cgreg = TRUE;
+		break;
+	default:
+		break;
+	}
+
+	do_register (NM_GSM_DEVICE (device));
+}
+
 static void
 power_up_response (NMSerialDevice *device,
                    int reply_index,
                    const char *reply,
                    gpointer user_data)
 {
-	/* Ignore errors */
-	do_register (NM_GSM_DEVICE (device));
+	NMGsmDevice *self = NM_GSM_DEVICE (device);
+	NMGsmDevicePrivate *priv = NM_GSM_DEVICE_GET_PRIVATE (self);
+	const char *responses[] = { "E160G", NULL };
+	const char *terminators[] = { "OK", "ERROR", "ERR", NULL };
+
+	/* Get the model the first time */
+	if (!priv->checked_cgmm)
+		modem_wait_for_reply (self, "AT+CGMM", 5, responses, terminators, get_model_done, NULL);
+	else
+		do_register (NM_GSM_DEVICE (device));
 }
 
 static void
@@ -496,13 +545,14 @@ init_full_done (NMSerialDevice *device,
 static void
 init_modem_full (NMGsmDevice *device)
 {
+	NMGsmDevicePrivate *priv = NM_GSM_DEVICE_GET_PRIVATE (device);
 	const char *responses[] = { "OK", "ERROR", "ERR", NULL };
 
 	/* Send E0 too because some devices turn echo back on after CPIN which
 	 * just breaks stuff since echo-ed commands are interpreted as replies.
 	 * rh #456770
 	 */
-	modem_wait_for_reply (device, "ATZ E0 V1 X4 &C1 +FCLASS=0", 10, responses, responses, init_full_done, NULL);
+	modem_wait_for_reply (device, modem_init_sequences[priv->init_tries], 10, responses, responses, init_full_done, NULL);
 }
 
 static void
@@ -647,7 +697,6 @@ init_done (NMSerialDevice *device,
 
 	switch (reply_index) {
 	case 0:
-		priv->init_tries = 0;
 		check_pin (NM_GSM_DEVICE (device));
 		break;
 	case -1:
@@ -657,35 +706,27 @@ init_done (NMSerialDevice *device,
 		                         NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED);
 		break;
 	default:
-		switch (priv->init_tries) {
-		case 0:
-			nm_warning ("Trying alternate modem initialization");
-			init_modem (device, "ATZ E0 V1 &C1");
-			break;
-		case 1:
-			nm_warning ("Trying second alternate modem initialization");
-			init_modem (device, "AT&F E0 V1");
-			break;
-		default:
+		priv->init_tries++;
+		if (modem_init_sequences[priv->init_tries] != NULL) {
+			nm_warning ("Trying alternate modem initialization (%d)",
+			            priv->init_tries);
+			init_modem (device);
+		} else {
 			nm_warning ("Modem initialization failed");
 			nm_device_state_changed (NM_DEVICE (device),
 			                         NM_DEVICE_STATE_FAILED,
 			                         NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED);
-			break;
 		}
-		priv->init_tries++;
-		return;
+		break;
 	}
 }
 
 static void
-init_modem (NMSerialDevice *device, gpointer user_data)
+init_modem (NMSerialDevice *device)
 {
+	NMGsmDevicePrivate *priv = NM_GSM_DEVICE_GET_PRIVATE (device);
 	const char *responses[] = { "OK", "ERROR", "ERR", NULL };
-	const char *init_string = user_data;
-
-	if (!init_string)
-		init_string = "ATZ E0 V1 X4 &C1 +FCLASS=0";
+	const char *init_string = modem_init_sequences[priv->init_tries];
 
 	modem_wait_for_reply (NM_GSM_DEVICE (device), init_string, 10, responses, responses, init_done, NULL);
 }
@@ -706,7 +747,7 @@ real_act_stage1_prepare (NMDevice *device, NMDeviceStateReason *reason)
 
 	NM_GSM_DEVICE_GET_PRIVATE (device)->init_tries = 0;
 
-	id = nm_serial_device_flash (serial_device, 100, init_modem, NULL);
+	id = nm_serial_device_flash (serial_device, 100, (NMSerialFlashFn) init_modem, NULL);
 	if (!id)
 		*reason = NM_DEVICE_STATE_REASON_UNKNOWN;
 
@@ -874,10 +915,8 @@ setup_monitor_device (NMGsmDevice *device)
 	GIOChannel *channel;
 	NMSettingSerial *setting;
 
-	if (!priv->monitor_iface) {
-		nm_debug ("No monitoring udi provided");
+	if (!priv->monitor_iface)
 		return FALSE;
-	}
 
 	priv->monitor_device = g_object_new (NM_TYPE_SERIAL_DEVICE,
 								  NM_DEVICE_INTERFACE_UDI, nm_device_get_udi (NM_DEVICE (device)),
