@@ -15,7 +15,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2008 Red Hat, Inc.
+ * Copyright (C) 2008 - 2009 Red Hat, Inc.
  */
 
 #include <stdlib.h>
@@ -45,16 +45,20 @@
 #include <nm-setting-ip4-config.h>
 #include <nm-setting-wired.h>
 #include <nm-setting-wireless.h>
+#include <nm-setting-8021x.h>
 #include <nm-utils.h>
 
 #include "common.h"
 #include "shvar.h"
+#include "sha1.h"
 
 #include "reader.h"
-#include "nm-system-config-interface.h"
 
-#define TYPE_ETHERNET "Ethernet"
-#define TYPE_WIRELESS "Wireless"
+#define PLUGIN_PRINT(pname, fmt, args...) \
+	{ g_message ("   " pname ": " fmt, ##args); }
+
+#define PLUGIN_WARN(pname, fmt, args...) \
+	{ g_warning ("   " pname ": " fmt, ##args); }
 
 static char *
 get_ifcfg_name (const char *file)
@@ -91,7 +95,7 @@ make_connection_setting (const char *file,
 {
 	NMSettingConnection *s_con;
 	char *ifcfg_name = NULL;
-	char *new_id = NULL, *uuid = NULL;
+	char *new_id = NULL, *uuid = NULL, *value;
 
 	ifcfg_name = get_ifcfg_name (file);
 	if (!ifcfg_name)
@@ -116,7 +120,12 @@ make_connection_setting (const char *file,
 
 	g_free (new_id);
 
-	uuid = nm_utils_uuid_generate_from_string (ifcfg->fileName);
+	/* Try for a UUID key before falling back to hashing the file name */
+	uuid = svGetValue (ifcfg, "UUID", FALSE);
+	if (!uuid || !strlen (uuid)) {
+		g_free (uuid);
+		uuid = nm_utils_uuid_generate_from_string (ifcfg->fileName);
+	}
 	g_object_set (s_con,
 	              NM_SETTING_CONNECTION_TYPE, type,
 	              NM_SETTING_CONNECTION_UUID, uuid,
@@ -124,8 +133,22 @@ make_connection_setting (const char *file,
 	g_free (uuid);
 
 	/* Be somewhat conservative about autoconnect */
-	if (svTrueValue (ifcfg, "ONBOOT", FALSE))
-		g_object_set (s_con, NM_SETTING_CONNECTION_AUTOCONNECT, TRUE, NULL);
+	g_object_set (s_con, NM_SETTING_CONNECTION_AUTOCONNECT,
+	              svTrueValue (ifcfg, "ONBOOT", FALSE),
+	              NULL);
+
+	value = svGetValue (ifcfg, "LAST_CONNECT", FALSE);
+	if (value) {
+		unsigned long int tmp;
+
+		errno = 0;
+		tmp = strtoul (value, NULL, 10);
+		if (errno == 0)
+			g_object_set (s_con, NM_SETTING_CONNECTION_TIMESTAMP, tmp, NULL);
+		else
+			PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: invalid LAST_CONNECT time");
+		g_free (value);
+	}
 
 	g_free (ifcfg_name);
 	return NM_SETTING (s_con);
@@ -143,10 +166,11 @@ get_one_ip4_addr (shvarFile *ifcfg,
 	g_return_if_fail (ifcfg != NULL);
 	g_return_if_fail (tag != NULL);
 	g_return_if_fail (out_addr != NULL);
+	g_return_if_fail (*out_addr == 0);
 	g_return_if_fail (error != NULL);
 	g_return_if_fail (*error == NULL);
 
-	value = svGetValue (ifcfg, tag);
+	value = svGetValue (ifcfg, tag, FALSE);
 	if (!value)
 		return;
 
@@ -167,42 +191,86 @@ get_one_ip4_addr (shvarFile *ifcfg,
 			goto error; \
 		if (dns) { \
 			if (!nm_setting_ip4_config_add_dns (s_ip4, dns)) \
-				PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS server %s", tag); \
+				PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS server %s", tag); \
 		} \
 	}
 		
 
 static NMSetting *
-make_ip4_setting (shvarFile *ifcfg, GError **error)
+make_ip4_setting (shvarFile *ifcfg, const char *network_file, GError **error)
 {
 	NMSettingIP4Config *s_ip4 = NULL;
 	char *value = NULL;
 	NMIP4Address *addr = NULL;
 	char *method = NM_SETTING_IP4_CONFIG_METHOD_MANUAL;
 	guint32 netmask = 0, tmp = 0;
+	shvarFile *network_ifcfg;
+	gboolean never_default = FALSE;
 
-	value = svGetValue (ifcfg, "BOOTPROTO");
-	if (value && (!g_ascii_strcasecmp (value, "bootp") || !g_ascii_strcasecmp (value, "dhcp")))
-		method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
+	network_ifcfg = svNewFile (network_file);
+	if (network_ifcfg) {
+		char *gatewaydev;
 
-	if (value && !g_ascii_strcasecmp (value, "autoip")) {
+		/* Get the connection ifcfg device name and the global gateway device */
+		value = svGetValue (ifcfg, "DEVICE", FALSE);
+		gatewaydev = svGetValue (network_ifcfg, "GATEWAYDEV", FALSE);
+
+		/* If there was a global gateway device specified, then only connections
+		 * for that device can be the default connection.
+		 */
+		if (gatewaydev && value && strcmp (value, gatewaydev))
+			never_default = TRUE;
+
+		g_free (gatewaydev);
 		g_free (value);
-		s_ip4 = (NMSettingIP4Config *) nm_setting_ip4_config_new ();
-		g_object_set (s_ip4, NM_SETTING_IP4_CONFIG_METHOD, NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL, NULL);
-		return NM_SETTING (s_ip4);
+		svCloseFile (network_ifcfg);
 	}
 
-	g_free (value);
+	value = svGetValue (ifcfg, "BOOTPROTO", FALSE);
+	if (value) {
+		if (!g_ascii_strcasecmp (value, "bootp") || !g_ascii_strcasecmp (value, "dhcp"))
+			method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
+		else if (!g_ascii_strcasecmp (value, "autoip")) {
+			g_free (value);
+			s_ip4 = (NMSettingIP4Config *) nm_setting_ip4_config_new ();
+			g_object_set (s_ip4,
+			              NM_SETTING_IP4_CONFIG_METHOD, NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL,
+			              NM_SETTING_IP4_CONFIG_NEVER_DEFAULT, never_default,
+			              NULL);
+			return NM_SETTING (s_ip4);
+		}
+		g_free (value);
+	} else {
+		char *tmp_ip4, *tmp_prefix, *tmp_netmask;
+
+		/* If there is no BOOTPROTO, no IPADDR, no PREFIX, and no NETMASK,
+		 * assume DHCP is to be used.  Happens with minimal ifcfg files like:
+		 *
+		 * DEVICE=eth0
+		 * HWADDR=11:22:33:44:55:66
+		 *
+		 */
+		tmp_ip4 = svGetValue (ifcfg, "IPADDR", FALSE);
+		tmp_prefix = svGetValue (ifcfg, "PREFIX", FALSE);
+		tmp_netmask = svGetValue (ifcfg, "NETMASK", FALSE);
+		if (!tmp_ip4 && !tmp_prefix && !tmp_netmask)
+			method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
+		g_free (tmp_ip4);
+		g_free (tmp_prefix);
+		g_free (tmp_netmask);
+	}
 
 	/* Handle manual settings */
 	if (!strcmp (method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL)) {
 		addr = nm_ip4_address_new ();
 
+		tmp = 0;
 		get_one_ip4_addr (ifcfg, "IPADDR", &tmp, error);
 		if (*error)
 			goto error;
 		nm_ip4_address_set_address (addr, tmp);
 
+		tmp = 0;
 		get_one_ip4_addr (ifcfg, "GATEWAY", &tmp, error);
 		if (*error)
 			goto error;
@@ -210,19 +278,18 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 
 		/* If no gateway in the ifcfg, try /etc/sysconfig/network instead */
 		if (!nm_ip4_address_get_gateway (addr)) {
-			shvarFile *network;
-
-			network = svNewFile ("/etc/sysconfig/network");
-			if (network) {
-				get_one_ip4_addr (network, "GATEWAY", &tmp, error);
-				svCloseFile (network);
+			network_ifcfg = svNewFile (network_file);
+			if (network_ifcfg) {
+				tmp = 0;
+				get_one_ip4_addr (network_ifcfg, "GATEWAY", &tmp, error);
+				svCloseFile (network_ifcfg);
 				if (*error)
 					goto error;
 				nm_ip4_address_set_gateway (addr, tmp);
 			}
 		}
 
-		value = svGetValue (ifcfg, "PREFIX");
+		value = svGetValue (ifcfg, "PREFIX", FALSE);
 		if (value) {
 			long int prefix;
 
@@ -240,6 +307,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 
 		/* Fall back to NETMASK if no PREFIX was specified */
 		if (!nm_ip4_address_get_prefix (addr)) {
+			netmask = 0;
 			get_one_ip4_addr (ifcfg, "NETMASK", &netmask, error);
 			if (*error)
 				goto error;
@@ -261,11 +329,12 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 	g_object_set (s_ip4,
 	              NM_SETTING_IP4_CONFIG_METHOD, method,
 	              NM_SETTING_IP4_CONFIG_IGNORE_AUTO_DNS, !svTrueValue (ifcfg, "PEERDNS", 1),
+	              NM_SETTING_IP4_CONFIG_NEVER_DEFAULT, never_default,
 	              NULL);
 
 	/* DHCP hostname for 'send host-name' option */
 	if (!strcmp (method, NM_SETTING_IP4_CONFIG_METHOD_AUTO)) {
-		value = svGetValue (ifcfg, "DHCP_HOSTNAME");
+		value = svGetValue (ifcfg, "DHCP_HOSTNAME", FALSE);
 		if (value && strlen (value))
 			g_object_set (s_ip4, NM_SETTING_IP4_CONFIG_DHCP_HOSTNAME, value, NULL);
 		g_free (value);
@@ -273,7 +342,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 
 	if (addr) {
 		if (!nm_setting_ip4_config_add_address (s_ip4, addr))
-			PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "    warning: duplicate IP4 address");
+			PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: duplicate IP4 address");
 	}
 
 	GET_ONE_DNS("DNS1");
@@ -281,7 +350,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 	GET_ONE_DNS("DNS3");
 
 	/* DNS searches */
-	value = svGetValue (ifcfg, "DOMAIN");
+	value = svGetValue (ifcfg, "DOMAIN", FALSE);
 	if (value) {
 		char **searches = NULL;
 
@@ -291,7 +360,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 			for (item = searches; *item; item++) {
 				if (strlen (*item)) {
 					if (!nm_setting_ip4_config_add_dns_search (s_ip4, *item))
-						PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS domain '%s'", *item);
+						PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS domain '%s'", *item);
 				}
 			}
 			g_strfreev (searches);
@@ -301,7 +370,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 
 	/* Legacy value NM used for a while but is incorrect (rh #459370) */
 	if (!nm_setting_ip4_config_get_num_dns_searches (s_ip4)) {
-		value = svGetValue (ifcfg, "SEARCH");
+		value = svGetValue (ifcfg, "SEARCH", FALSE);
 		if (value) {
 			char **searches = NULL;
 
@@ -311,7 +380,7 @@ make_ip4_setting (shvarFile *ifcfg, GError **error)
 				for (item = searches; *item; item++) {
 					if (strlen (*item)) {
 						if (!nm_setting_ip4_config_add_dns_search (s_ip4, *item))
-							PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS search '%s'", *item);
+							PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: duplicate DNS search '%s'", *item);
 					}
 				}
 				g_strfreev (searches);
@@ -378,7 +447,7 @@ read_mac_address (shvarFile *ifcfg, GByteArray **array, GError **error)
 	g_return_val_if_fail (error != NULL, FALSE);
 	g_return_val_if_fail (*error == NULL, FALSE);
 
-	value = svGetValue (ifcfg, "HWADDR");
+	value = svGetValue (ifcfg, "HWADDR", FALSE);
 	if (!value || !strlen (value)) {
 		g_free (value);
 		return TRUE;
@@ -414,7 +483,7 @@ add_one_wep_key (shvarFile *ifcfg,
 	g_return_val_if_fail (key_idx <= 3, FALSE);
 	g_return_val_if_fail (s_wsec != NULL, FALSE);
 
-	value = svGetValue (ifcfg, shvar_key);
+	value = svGetValue (ifcfg, shvar_key, FALSE);
 	if (!value || !strlen (value)) {
 		g_free (value);
 		return TRUE;
@@ -522,9 +591,9 @@ read_wep_keys (shvarFile *ifcfg,
 }
 
 static NMSetting *
-make_wireless_security_setting (shvarFile *ifcfg,
-                                const char *file,
-                                GError **error)
+make_wep_setting (shvarFile *ifcfg,
+                  const char *file,
+                  GError **error)
 {
 	NMSettingWirelessSecurity *s_wireless_sec;
 	char *value;
@@ -532,8 +601,9 @@ make_wireless_security_setting (shvarFile *ifcfg,
 	int default_key_idx = 0;
 
 	s_wireless_sec = NM_SETTING_WIRELESS_SECURITY (nm_setting_wireless_security_new ());
+	g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, "none", NULL);
 
-	value = svGetValue (ifcfg, "DEFAULTKEY");
+	value = svGetValue (ifcfg, "DEFAULTKEY", FALSE);
 	if (value) {
 		gboolean success;
 
@@ -579,7 +649,7 @@ make_wireless_security_setting (shvarFile *ifcfg,
 		goto error;
 	}
 
-	value = svGetValue (ifcfg, "SECURITYMODE");
+	value = svGetValue (ifcfg, "SECURITYMODE", FALSE);
 	if (value) {
 		char *lcase;
 
@@ -618,9 +688,6 @@ make_wireless_security_setting (shvarFile *ifcfg,
 		/* Unencrypted */
 		g_object_unref (s_wireless_sec);
 		s_wireless_sec = NULL;
-	} else {
-		// FIXME: WEP-only for now
-		g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, "none", NULL);
 	}
 
 	return (NMSetting *) s_wireless_sec;
@@ -631,10 +698,257 @@ error:
 	return NULL;
 }
 
+static gboolean
+fill_wpa_ciphers (shvarFile *ifcfg,
+                  NMSettingWirelessSecurity *wsec,
+                  gboolean group,
+                  gboolean adhoc)
+{
+	char *value = NULL, *p;
+	char **list = NULL, **iter;
+	int i = 0;
+
+	p = value = svGetValue (ifcfg, group ? "CIPHER_GROUP" : "CIPHER_PAIRWISE", TRUE);
+	if (!value)
+		return TRUE;
+
+	/* Strip quotes */
+	if (p[0] == '"')
+		p++;
+	if (p[strlen (p) - 1] == '"')
+		p[strlen (p) - 1] = '\0';
+
+	list = g_strsplit_set (p, " ", 0);
+	for (iter = list; iter && *iter; iter++, i++) {
+		/* Ad-Hoc configurations cannot have pairwise ciphers, and can only
+		 * have one group cipher.  Ignore any additional group ciphers and
+		 * any pairwise ciphers specified.
+		 */
+		if (adhoc) {
+			if (group && (i > 0)) {
+				PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: ignoring group cipher '%s' (only one group cipher allowed in Ad-Hoc mode)",
+				             *iter);
+				continue;
+			} else if (!group) {
+				PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: ignoring pairwise cipher '%s' (pairwise not used in Ad-Hoc mode)",
+				             *iter);
+				continue;
+			}
+		}
+
+		if (!strcmp (*iter, "CCMP")) {
+			if (group)
+				nm_setting_wireless_security_add_group (wsec, "ccmp");
+			else
+				nm_setting_wireless_security_add_pairwise (wsec, "ccmp");
+		} else if (!strcmp (*iter, "TKIP")) {
+			if (group)
+				nm_setting_wireless_security_add_group (wsec, "tkip");
+			else
+				nm_setting_wireless_security_add_pairwise (wsec, "tkip");
+		} else if (group && !strcmp (*iter, "WEP104"))
+			nm_setting_wireless_security_add_group (wsec, "wep104");
+		else if (group && !strcmp (*iter, "WEP40"))
+			nm_setting_wireless_security_add_group (wsec, "wep40");
+		else {
+			PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: ignoring invalid %s cipher '%s'",
+			             group ? "CIPHER_GROUP" : "CIPHER_PAIRWISE",
+			             *iter);
+		}
+	}
+
+	if (list)
+		g_strfreev (list);
+	g_free (value);
+	return TRUE;
+}
+
+#define WPA_PMK_LEN 32
+
+static char *
+parse_wpa_psk (shvarFile *ifcfg,
+               const char *file,
+               const GByteArray *ssid,
+               GError **error)
+{
+	shvarFile *keys_ifcfg;
+	char *psk = NULL, *p, *hashed = NULL;
+
+	/* Passphrase must be between 10 and 66 characters in length becuase WPA
+	 * hex keys are exactly 64 characters (no quoting), and WPA passphrases
+	 * are between 8 and 63 characters (inclusive), plus optional quoting if
+	 * the passphrase contains spaces.
+	 */
+
+	/* Try to get keys from the "shadow" key file */
+	keys_ifcfg = get_keys_ifcfg (file);
+	if (keys_ifcfg) {
+		psk = svGetValue (keys_ifcfg, "WPA_PSK", TRUE);
+		svCloseFile (keys_ifcfg);
+	}
+
+	/* Fall back to the original ifcfg */
+	if (!psk)
+		psk = svGetValue (ifcfg, "WPA_PSK", TRUE);
+
+	if (!psk) {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "Missing WPA_PSK for WPA-PSK key management");
+		return NULL;
+	}
+
+	p = psk;
+	if (p[0] == '"' && psk[strlen (psk) - 1] == '"') {
+		unsigned char *buf;
+
+		/* Get rid of the quotes */
+		p++;
+		p[strlen (p) - 1] = '\0';
+
+		/* Length check */
+		if (strlen (p) < 8 || strlen (p) > 63) {
+			g_set_error (error, ifcfg_plugin_error_quark (), 0,
+			             "Invalid WPA_PSK (passphrases must be between "
+			             "8 and 63 characters long (inclusive))");
+			goto out;
+		}
+
+		/* hash the passphrase to a hex key */
+		buf = g_malloc0 (WPA_PMK_LEN * 2);
+		pbkdf2_sha1 (p, (char *) ssid->data, ssid->len, 4096, buf, WPA_PMK_LEN);
+		hashed = utils_bin2hexstr ((const char *) buf, WPA_PMK_LEN, WPA_PMK_LEN * 2);
+		g_free (buf);
+	} else if (strlen (psk) == 64) {
+		/* Verify the hex PSK; 64 digits */
+		while (*p) {
+			if (!isxdigit (*p++)) {
+				g_set_error (error, ifcfg_plugin_error_quark (), 0,
+				             "Invalid WPA_PSK (contains non-hexadecimal characters)");
+				goto out;
+			}
+		}
+		hashed = g_strdup (psk);
+	} else {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "Invalid WPA_PSK (doesn't look like a passphrase or hex key)");
+		goto out;
+	}
+
+out:
+	g_free (psk);
+	return hashed;
+}
+
+#if 0
+static NMSetting8021x *
+fill_8021x (shvarFile *ifcfg,
+            const char *file,
+            const char *key_mgmt,
+            GError **error)
+{
+	return NULL;
+}
+#endif
+
+static NMSetting *
+make_wpa_setting (shvarFile *ifcfg,
+                  const char *file,
+                  const GByteArray *ssid,
+                  gboolean adhoc,
+                  NMSetting8021x **s_8021x,
+                  GError **error)
+{
+	NMSettingWirelessSecurity *wsec;
+	char *value, *psk;
+
+	wsec = NM_SETTING_WIRELESS_SECURITY (nm_setting_wireless_security_new ());
+
+	value = svGetValue (ifcfg, "KEY_MGMT", FALSE);
+	if (!value)
+		goto error; /* Not WPA or Dynamic WEP */
+
+	/* Pairwise and Group ciphers */
+	fill_wpa_ciphers (ifcfg, wsec, FALSE, adhoc);
+	fill_wpa_ciphers (ifcfg, wsec, TRUE, adhoc);
+
+	/* WPA and/or RSN */
+	if (adhoc) {
+		/* Ad-Hoc mode only supports WPA proto for now */
+		nm_setting_wireless_security_add_proto (wsec, "wpa");
+	} else {
+		if (svTrueValue (ifcfg, "WPA_ALLOW_WPA", TRUE))
+			nm_setting_wireless_security_add_proto (wsec, "wpa");
+		if (svTrueValue (ifcfg, "WPA_ALLOW_WPA2", TRUE))
+			nm_setting_wireless_security_add_proto (wsec, "rsn");
+	}
+
+	if (!strcmp (value, "WPA-PSK")) {
+		psk = parse_wpa_psk (ifcfg, file, ssid, error);
+		if (!psk)
+			goto error;
+		g_object_set (wsec, NM_SETTING_WIRELESS_SECURITY_PSK, psk, NULL);
+		g_free (psk);
+
+		if (adhoc)
+			g_object_set (wsec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, "wpa-none", NULL);
+		else
+			g_object_set (wsec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, "wpa-psk", NULL);
+#if 0
+	} else if (!strcmp (value, "WPA-EAP") || !strcmp (value, "IEEE8021X")) {
+		/* Adhoc mode is mutually exclusive with any 802.1x-based authentication */
+		if (adhoc) {
+			g_set_error (error, ifcfg_plugin_error_quark (), 0,
+			             "Ad-Hoc mode cannot be used with KEY_MGMT type '%s'", value);
+			goto error;
+		}
+
+		*s_8021x = fill_8021x (ifcfg, file, value, error);
+		if (!*s_8021x)
+			goto error;
+#endif
+	} else {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "Unknown wireless KEY_MGMT type '%s'", value);
+		goto error;
+	}
+
+	g_free (value);
+	return (NMSetting *) wsec;
+
+error:
+	g_free (value);
+	if (wsec)
+		g_object_unref (wsec);
+	return NULL;
+}
+
+static NMSetting *
+make_wireless_security_setting (shvarFile *ifcfg,
+                                const char *file,
+                                const GByteArray *ssid,
+                                gboolean adhoc,
+                                NMSetting8021x **s_8021x,
+                                GError **error)
+{
+	NMSetting *wsec;
+
+	wsec = make_wpa_setting (ifcfg, file, ssid, adhoc, s_8021x, error);
+	if (wsec)
+		return wsec;
+	else if (*error)
+		return NULL;
+
+	wsec = make_wep_setting (ifcfg, file, error);
+	if (wsec)
+		return wsec;
+	else if (*error)
+		return NULL;
+
+	return NULL; /* unencrypted */
+}
 
 static NMSetting *
 make_wireless_setting (shvarFile *ifcfg,
-                       NMSetting *security,
                        gboolean unmanaged,
                        GError **error)
 {
@@ -652,7 +966,7 @@ make_wireless_setting (shvarFile *ifcfg,
 		return NULL;
 	}
 
-	value = svGetValue (ifcfg, "ESSID");
+	value = svGetValue (ifcfg, "ESSID", FALSE);
 	if (value) {
 		gsize len = strlen (value);
 
@@ -678,7 +992,7 @@ make_wireless_setting (shvarFile *ifcfg,
 	}
 
 	if (!unmanaged) {
-		value = svGetValue (ifcfg, "MODE");
+		value = svGetValue (ifcfg, "MODE", FALSE);
 		if (value) {
 			char *lcase;
 			const char *mode = NULL;
@@ -692,7 +1006,7 @@ make_wireless_setting (shvarFile *ifcfg,
 				mode = "infrastructure";
 			} else {
 				g_set_error (error, ifcfg_plugin_error_quark (), 0,
-				             "Invalid mode '%s' (not ad-hoc or managed)",
+				             "Invalid mode '%s' (not 'Ad-Hoc' or 'Managed')",
 				             lcase);
 				g_free (lcase);
 				goto error;
@@ -701,11 +1015,6 @@ make_wireless_setting (shvarFile *ifcfg,
 
 			g_object_set (s_wireless, NM_SETTING_WIRELESS_MODE, mode, NULL);
 		}
-
-		if (security)
-			g_object_set (s_wireless, NM_SETTING_WIRELESS_SEC,
-						  NM_SETTING_WIRELESS_SECURITY_SETTING_NAME, NULL);
-
 		// FIXME: channel/freq, other L2 parameters like RTS
 	}
 
@@ -726,9 +1035,12 @@ wireless_connection_from_ifcfg (const char *file,
 	NMConnection *connection = NULL;
 	NMSetting *con_setting = NULL;
 	NMSetting *wireless_setting = NULL;
+	NMSetting8021x *s_8021x = NULL;
 	const GByteArray *ssid;
 	NMSetting *security_setting = NULL;
 	char *printable_ssid = NULL;
+	const char *mode;
+	gboolean adhoc = FALSE;
 
 	g_return_val_if_fail (file != NULL, NULL);
 	g_return_val_if_fail (ifcfg != NULL, NULL);
@@ -742,17 +1054,8 @@ wireless_connection_from_ifcfg (const char *file,
 		return NULL;
 	}
 
-	/* Wireless security */
-	security_setting = make_wireless_security_setting (ifcfg, file, error);
-	if (*error) {
-		g_object_unref (connection);
-		return NULL;
-	}
-	if (security_setting)
-		nm_connection_add_setting (connection, security_setting);
-
 	/* Wireless */
-	wireless_setting = make_wireless_setting (ifcfg, security_setting, unmanaged, error);
+	wireless_setting = make_wireless_setting (ifcfg, unmanaged, error);
 	if (!wireless_setting) {
 		g_object_unref (connection);
 		return NULL;
@@ -765,6 +1068,28 @@ wireless_connection_from_ifcfg (const char *file,
 	else
 		printable_ssid = g_strdup_printf ("unmanaged");
 
+	if (!unmanaged) {
+		mode = nm_setting_wireless_get_mode (NM_SETTING_WIRELESS (wireless_setting));
+		if (mode && !strcmp (mode, "adhoc"))
+			adhoc = TRUE;
+
+		/* Wireless security */
+		security_setting = make_wireless_security_setting (ifcfg, file, ssid, adhoc, &s_8021x, error);
+		if (*error) {
+			g_object_unref (connection);
+			return NULL;
+		}
+		if (security_setting) {
+			nm_connection_add_setting (connection, security_setting);
+			if (s_8021x)
+				nm_connection_add_setting (connection, NM_SETTING (s_8021x));
+
+			g_object_set (wireless_setting, NM_SETTING_WIRELESS_SEC,
+			              NM_SETTING_WIRELESS_SECURITY_SETTING_NAME, NULL);
+		}
+	}
+
+	/* Connection */
 	con_setting = make_connection_setting (file, ifcfg,
 	                                       NM_SETTING_WIRELESS_SETTING_NAME,
 	                                       printable_ssid);
@@ -798,14 +1123,14 @@ make_wired_setting (shvarFile *ifcfg, gboolean unmanaged, GError **error)
 
 	s_wired = NM_SETTING_WIRED (nm_setting_wired_new ());
 
-	value = svGetValue (ifcfg, "MTU");
+	value = svGetValue (ifcfg, "MTU", FALSE);
 	if (value) {
 		if (get_int (value, &mtu)) {
 			if (mtu >= 0 && mtu < 65536)
 				g_object_set (s_wired, NM_SETTING_WIRED_MTU, mtu, NULL);
 		} else {
 			/* Shouldn't be fatal... */
-			PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "    warning: invalid MTU '%s'", value);
+			PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: invalid MTU '%s'", value);
 		}
 		g_free (value);
 	}
@@ -908,9 +1233,12 @@ is_wireless_device (const char *iface)
 
 NMConnection *
 connection_from_file (const char *filename,
+                      const char *network_file,
+                      const char *test_type,  /* for unit tests only */
                       gboolean *ignored,
                       char **keyfile,
-                      GError **error)
+                      GError **error,
+                      gboolean *ignore_error)
 {
 	NMConnection *connection = NULL;
 	NMSettingConnection *s_con;
@@ -925,6 +1253,10 @@ connection_from_file (const char *filename,
 	g_return_val_if_fail (keyfile != NULL, NULL);
 	g_return_val_if_fail (*keyfile == NULL, NULL);
 
+	/* Non-NULL only for unit tests; normally use /etc/sysconfig/network */
+	if (!network_file)
+		network_file = SYSCONFDIR "/sysconfig/network";
+
 	ifcfg_name = get_ifcfg_name (filename);
 	if (!ifcfg_name) {
 		g_set_error (error, ifcfg_plugin_error_quark (), 0,
@@ -940,14 +1272,14 @@ connection_from_file (const char *filename,
 		return NULL;
 	}
 
-	type = svGetValue (parsed, "TYPE");
+	type = svGetValue (parsed, "TYPE", FALSE);
 	if (!type) {
 		char *device;
 
 		/* If no type, if the device has wireless extensions, it's wifi,
 		 * otherwise it's ethernet.
 		 */
-		device = svGetValue (parsed, "DEVICE");
+		device = svGetValue (parsed, "DEVICE", FALSE);
 		if (!device) {
 			g_set_error (error, ifcfg_plugin_error_quark (), 0,
 			             "File '%s' had neither TYPE nor DEVICE keys.", filename);
@@ -955,22 +1287,31 @@ connection_from_file (const char *filename,
 		}
 
 		if (!strcmp (device, "lo")) {
+			*ignore_error = TRUE;
 			g_set_error (error, ifcfg_plugin_error_quark (), 0,
 			             "Ignoring loopback device config.");
 			g_free (device);
 			goto done;
 		}
 
-		/* Test wireless extensions */
-		if (is_wireless_device (device))
-			type = g_strdup (TYPE_WIRELESS);
-		else
-			type = g_strdup (TYPE_ETHERNET);
+		if (!test_type) {
+			/* Test wireless extensions */
+			if (is_wireless_device (device))
+				type = g_strdup (TYPE_WIRELESS);
+			else
+				type = g_strdup (TYPE_ETHERNET);
+		} else {
+			/* For the unit tests, there won't necessarily be any
+			 * adapters of the connection's type in the system so the
+			 * type can't be tested with ioctls.
+			 */
+			type = g_strdup (test_type);
+		}
 
 		g_free (device);
 	}
 
-	nmc = svGetValue (parsed, "NM_CONTROLLED");
+	nmc = svGetValue (parsed, "NM_CONTROLLED", FALSE);
 	if (nmc) {
 		char *lower;
 
@@ -1004,7 +1345,7 @@ connection_from_file (const char *filename,
 	if (!connection || *ignored)
 		goto done;
 
-	s_ip4 = make_ip4_setting (parsed, error);
+	s_ip4 = make_ip4_setting (parsed, network_file, error);
 	if (*error) {
 		g_object_unref (connection);
 		connection = NULL;
